/** 
 *  j6502.java
 *  Richard Russo
 * 
 *  6502 emulator.  The entire documented instruction set is 
 *  implemented, as well as all addressing modes, BCD mode,
 *  etc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

import java.lang.Integer;
import java.text.Format;
import java.text.NumberFormat;

public class j6502 {
    //variables to represent the 6502 registers
    //the following registers are 8-bit

    public static final int CARRY = 0, ZERO = 1, INTERRUPT = 2, DECIMAL = 3, 
	BREAK = 4, OVERFLOW = 6, SIGN = 7;

    public static int MEM_SIZE = 65536;



    public byte a = 0;                       //accumulator
    public byte x = 0;                       //x index register
    public byte y = 0;                       //y index register
    public byte sp = (byte)0xFF;             //stack pointer
    public byte p = 0;                       //status register
    
    //flags represented as boolean values
    boolean[] flags = {false, false, true, false, false, false, false, false};
    
    public int pc = 0;  //program counter (16-bit register)
    
    //64k to represent the memory that the 6502 can address
    public byte[] memory;
    
    
    public j6502() {

	reset ();
	memory = new byte[65536];
    }

    public j6502 (int ip)
    {
	reset ();
	pc = ip;
	memory = new byte [65536];
    }

    public j6502 (byte[] b)
    {
	reset ();
	memory = b;
    }

	

    public void reset ()
    {
	flags[CARRY]     = false;
	flags[ZERO]      = false;
	flags[INTERRUPT] = true;
	flags[DECIMAL]   = false;
	flags[BREAK]     = false;
	flags[5]         = false;
	flags[OVERFLOW]  = false;
	flags[SIGN]      = false;
	
	a = 0;
	x = y = 0;
	sp = (byte)0xFF;
	p = calc_status ();
	pc = 0;
    }


    public void screen_dump ()
    {
	int count;
	byte regs[] = { this.a , this.x, 
		       this.y, this.sp };
	String[] names = { "A ", "X ", "Y ", "SP" };
	String col = "00000000";
	String tmp;

	System.out.println ("\nFlags:");
	System.out.println ("SV_BDIZC");
	tmp = Integer.toString (calc_status () & 0xFF, 2);
	System.out.println (col.substring (0, col.length () - tmp.length ()) + tmp);
	System.out.println ();

	for (count = 0; count < regs.length; count++)
	    {
		tmp = Integer.toString (regs[count] & 0xFF, 2);
		System.out.println 
		    (names[count] + ": " 
		     + col.substring (0, col.length () - tmp.length ()) + tmp
		     + "  0x" 
		     + Integer.toString (regs[count] & 0xFF, 16));
	    }

	System.out.println ("PC: " + Integer.toString (pc));

    }
    
    public static void main(String args[]) {
	j6502 cpu = new j6502 ();
	NumberFormat n = NumberFormat.getInstance ();
	int step = 0x20;

	int count;

        System.out.println("a Motorola 6502 emulator in java");

	System.out.println ("\nADC binary test:");

	cpu.a = 0;

	
	for (count = 0; count < 9; count++)
	    {
		cpu.adc ((byte)step);

		cpu.screen_dump ();

		System.out.println ();
	    }

	System.out.println ("\n\nADC decimal test:");
	cpu.reset ();

	for (count = 0; count < 9; count++)
	    {
		cpu.adc_bcd ((byte)step);
		cpu.screen_dump ();

		System.out.println ();
	    }

	System.out.println ("test 2");

	cpu.reset ();
	cpu.a = (byte)(0x68);
	cpu.adc_bcd ((byte)0x22);
	cpu.screen_dump ();
    }
    
    public byte calc_status () {
	int count;
	byte val = 0;
	for(count = 7; count>=0; count--) {
	    val <<= 1;  //left-shift 1 bit
	    if (flags[count]) val++; //set least significant bit if flag is true
	}
	return val;
    }
    
    public void set_status (byte b) {
	int count;
	byte val = 1;

	for(count = 0; count < 8; count++) {

	    if ((b & val) != 0) flags[count] = true;
	    else flags[count] = false;

	    val <<= 1;
	}
    }
    
    private void push(byte b) {
	memory[0x100 + (sp & 0xFF)] = b;
	sp--;
    }
    
    private byte pop() {
	return memory[0x100 + ((++sp) & 0xFF)];
    }

    /**
     * Set the carry flag according to the most significant bit of the
     * byte b.
     *************************************************************/

    private void set_carry_msb (byte b)
    {
	if (b < 0)
	    {
		flags[CARRY] = true;
	    }
	else
	    {
		flags[CARRY] = false;
	    }
    }

    private void set_carry_lsb (byte b)
    {
	if ((b & 1) != 0)
	    {
		flags[CARRY] = true;
	    }
	else
	    {
		flags[CARRY] = false;
	    }
    }


    /* a - b */

    private void cmp (byte a, byte b)
    {
	int i;
	
	b--;
	b ^= 0xFF;
	
	i = (a & 0xFF) + (b & 0xFF);

	if ((i & 0x100) != 0) flags[CARRY] = true;
	else flags[CARRY] = false;

	set_ar_flags ((byte)(i & 0xFF));
    }

	

    /**
     * Set the sign flag based on the msb of b.
     *************************************************************/

    private void set_sign (byte b)
    {
	if (b < 0)
	    {
		flags[SIGN] = true;
	    }
	else
	    {
		flags[SIGN] = false;
	    }
    }

    public void adc (byte b)
    {
	int result;
	int carry6;

	result = (b & 0xFF) + (this.a & 0xFF);
	carry6 = (b & 0x7F) + (this.a & 0x7F);

	if (flags[CARRY]) {
	    result++;
	    carry6++;
	}

	flags[CARRY] = ((result & 0x100) != 0);

	flags[OVERFLOW] = flags[CARRY] ^ ((carry6 & 0x80) != 0);

	this.a = (byte)(result & 0xFF);

	set_ar_flags (this.a);
    }


    void adc_bcd (byte b)
    {
	int low, high;
	int result;

	low = (a & 0x0F) + (b & 0x0F);
	if (flags[CARRY])
	    {
		low++;
	    }

	high = (a & 0xF0) + (b & 0xF0);


	flags[ZERO] = ((low + high) & 0xFF) == 0;
	
	if (low > 9) {
	    low += 6;
	}




	flags[SIGN] = (((low + high) & 0x80) != 0);
	flags[OVERFLOW] = ((((low + ((a & 0x70) + (b & 0x70))) & 0x80) << 1)   //bit 6 carry
			   ^ ((low + high) & 0x100)) != 0;                     //bit 7 carry
	
	if (high > 0x90) {
	    high += 0x60;
	}

	result = high + low;
	this.a = (byte)(result);
	flags[CARRY] = ((result & 0x100) != 0);
    }

    public void sbc (byte b)
    {
	b--;
	b ^= 0xFF;
	adc (b);
    }

    public void sbc_bcd (byte b)
    {
	int low, high;
	int result;

	low = (a & 0x0F) - (b & 0x0F);
	
	if (!flags[CARRY])
	    {
		low --;
	    }

	low &= 0xFF;
	if ((low & 0x10) != 0)
	    {
		low -= 6;
	    }

	high = (a & 0xF0) - (b & 0xF0);
	high &= 0xFF;

	if ((low & 0x10) != 0)
	    {
		high -= 0x10;
	    }

	if ((high & 0x100) != 0)
	    {
		high -= 0x60;
	    }

	result = (a & 0xFF) - (b & 0xFF);
	if (flags[CARRY])
	    {
		result --;
	    }

	flags[CARRY]    = (result & 0x100) != 0;
	flags[SIGN]     = (result & 0x80) != 0;
	flags[ZERO]     = (result & 0xFF) == 0;
	flags[OVERFLOW] = (((result ^ a) & 0x80) != 0) && (((a ^ b) & 0x80) != 0);

	a = (byte)((high & 0xF0) + (low & 0x0F));

    }


    /**
     * Set the arithmetic flags (ZERO, SIGN) according
     * to the contents of b.
     *************************************************************/

    private void set_ar_flags (byte b)
    {
	if (b == 0)
	    {
		flags[ZERO] = true;
		flags[SIGN] = false;
	    }
	else
	    {
		flags[ZERO] = false;
		if (b < 0)
		    {
			flags[SIGN] = true;
		    }
	    }
    }


    /**
     * The 6502 addressing modes:
     *
     * 1) Immediate
     * 2) Direct, zero page (8-bit address)
     * 3) Direct, 16-bit address
     * 4) Pre-indexed indirect
     * 5) Post-indexed indirect
     * 6) Indexed by x zero page
     * 7) Indexed by x
     * 8) Indexed by y
     * 9) Indirect
     *
     * The last few addressing modes are implemented as needed.
     *
     * pc is incremented inside of these functions!!
     *************************************************************/

    private byte immediate ()
    {
	pc++;
	return memory[pc++];
    }
    
    private int direct_zp ()
    {
	int addr = 0;

	pc++;
	addr += (memory[pc] & 0xFF);
	pc++;
	return addr;
    }
    
    private int direct ()
    {
	int addr = 0;

	pc++;
	addr += (memory[pc] & 0xFF);
	pc++;
	addr += (memory[pc] & 0xFF) << 8;
	pc++;
	return addr;
    }

    private int pre_idx ()
    {
	int idx;
	int addr = 0;
    
	pc++;
	idx = (this.x & 0xFF) + (memory[pc] & 0xFF);
	idx %= 256;
	pc++;

	addr += memory[idx] & 0xFF;
	idx++;
	addr += (memory[idx] & 0xFF) << 8;


	return addr;
    }

    private int post_idx ()
    {
	int idx;
	int addr;

	pc++;
	idx = memory[pc] & 0xFF;
	pc++;

	addr = memory[idx] & 0xFF;
	idx++;
	addr += (memory[idx] & 0xFF) << 8;
	addr += this.y & 0xFF;

	if (addr > 65535)
	    {
		addr %= 65536;
	    }

	return addr;
    }
	
    private int idx_x_zp ()
    {
	int addr;

	pc++;
	addr = memory[pc] & 0xFF;
	pc++;
	
	addr += this.x & 0xFF;

	addr %= 256;
	
	return addr;
    }

    private int idx_y_zp ()
    {
	int addr;
	
	pc++;
	addr = memory[pc] & 0xFF;
	pc++;

	addr += (this.y & 0xFF);
	
	addr %= 256;

	return addr;
    }

    private int idx_x ()
    {
	int addr;

	pc++;
	addr = (memory[pc] & 0xFF);
	pc++;
	addr += (memory[pc] & 0xFF) << 8;
	pc++;
	
	addr += (this.x & 0xFF);
	
	if (addr > 65535)
	    {
		addr %= 65536;
	    }
	
	return addr;
    }

    private int idx_y ()
    {
	int addr;
	
	pc++;
	addr = (memory[pc] & 0xFF);
	pc++;
	addr += (memory[pc] & 0xFF) << 8;
	pc++;

	addr += (this.y & 0xFF);

	if (addr > 65535)
	    {
		addr %= 65536;
	    }

	return addr;
    }

    private int indirect ()
    {
	int addr, eff;

	pc++;
	addr = memory[pc] & 0xFF;
	pc++;
	addr += (memory[pc] & 0xFF) << 8;
	pc++;

	eff = memory[addr] & 0xFF;
	addr++;
	if (addr > 65535)
	    {
		addr %= 65536;
	    }
	eff = (memory[addr] & 0xFF) << 8;

	return eff;
    }

	
    //execute() - execute 1 instruction at pc
    public void execute () 
    {
	int addr = 0, i;
	byte temp;

	if (pc > 65535) {
	    pc %= 65536;
	}
	
	
	switch(memory[pc] & 0xFF) {
	case 0x00:
	    //BRK -- force break  [0x00]
	    //no addressing   [1 byte]
	    
	    
	    flags[INTERRUPT] = true;  //set interrupt mask flag - disable interrupts
	    flags[BREAK] = true;  //set break status flag
	    
	    pc++;
	    pc++; //increment program counter by 2
	    push((byte)(pc & 0xFF));  //push low byte of pc
	    push((byte)((pc >> 8) & 0xFF)); //push high byte of pc
	    push(calc_status ()); //push status byte
	    //load the pc register with the interrupt vector:
	    //low byte address 0xFFFE, high byte 0xFFFF
	    pc = ((memory[0xFFFF] << 8) + memory[0xFFFE]) & 0xFFFF;
	    break;
	    
	case 0x01:
	    //ORA -- Logical OR: memory with accumulator [0x01]
	    //Pre-indexed indirect addressing mode:  ORA (addr,x)   [2 bytes]
	    
	    
	    a |= memory[pre_idx ()];
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x02:
	    break;
	case 0x03:
	    break;
	case 0x04:
	    break;
	    
	case 0x05:
	    //ORA -- Logical OR: memory with accumulator [0x05]
	    //Zero page direct addressing mode:   ORA addr  [2 bytes]
	    
	    
	    a |= memory[direct_zp ()];
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x06:
	    //ASL -- arithmetic shift left contents of memory [0x06]
	    //Zero page direct addressing mode:   ASL addr    [2 bytes]
	    
	    //if the most significant bit is set, shift it into the carry flag
	    
	    addr = direct_zp ();
	    
	    set_carry_msb (memory[addr]);
	    
	    memory[addr] <<= 1;
	    
	    set_ar_flags (memory[addr]);
	    
	    break;
	    
	case 0x07:
	    break;
	    
	case 0x08:
	    //PHP -- Push Status Register onto stack
	    //Implied addressing    [1 byte]
	    
	    push(calc_status ());
	    pc++;
	    break;
	    
	case 0x09:
	    //ORA -- Logical OR: Accumulator with data [0x09]
	    //Immediate addressing mode:  ORA data    [2 bytes]
	    
	    a |= immediate ();
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x0A:
	    //ASL -- Arithmetic shift left accumulator [0x0A]
	    //register operand: ASL A [1 byte]
	    
	    
	    //shift most significant bit into carry flag
	    
	    set_carry_msb (a);
	    
	    a <<= 1;
	    
	    set_ar_flags (a);
	    
	    pc++;
	    break;
	    
	case 0x0B:
	    break;
	    
	case 0x0C:
	    break;
	    
	case 0x0D:
	    //ORA -- Logical OR accumulator with memory  [0x0D]
	    //direct addressing mode:   ORA addr16    [3 bytes]
	    
	    
	    a |= memory[direct ()];
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x0E:
	    //ASL -- Arithmetic shift left memory contents   [0x0E]
	    //Direct addressing mode:   ASL addr16    [3 bytes]
	    
	    addr = direct ();
	    
	    set_carry_msb (memory[addr]);
	    
	    memory[addr] <<= 1;
	    
	    set_ar_flags (memory[addr]);
	    
	    break;
	    
	case 0x0F:
	    break;
	    
	case 0x10:
	    //BPL -- Branch if plus (sign flag is zero)   [0x10]
	    //Displacement operand:   BPL disp    [2 bytes]
	    
	    pc++;
	    if (!flags[7]) pc += memory[pc]; //displacement operand is treated as a signed byte
	    pc++;
	    break;
	    
	case 0x11:
	    //ORA -- Logical OR memory with accumulator [0x11]
	    //Post-indexed indirect addressing mode:    ORA (addr),Y    [2 bytes]
	    
	    a |= memory[post_idx ()];
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x12:
	    break;
	    
	case 0x13:
	    break;
	    
	case 0x14:
	    break;
	    
	case 0x15:
	    //ORA -- Logical OR  accumulator with memory  [0x15]
	    //Zero page indexed with X:   ORA addr,X    [2 bytes]
	    
	    a |= memory[idx_x_zp ()];
	    
	    set_ar_flags (a);
	    
	    break;
	    
	    
	case 0x16:
	    // ASL -- Arithmetic shift left   [0x16]
	    // Zero page indexed with X:    ASL addr,X    [2 bytes]
	    
	    addr = idx_x_zp ();
	    
	    set_carry_msb (memory[addr]);
	    
	    memory[addr] <<= 1;
	    
	    set_ar_flags (memory[addr]);
	    
	    break;
	    
	case 0x17:
	    break;
	    
	case 0x18:
	    //CLC -- Clear Carry Flag   [1 byte]
	    
	    flags[0] = false;
	    pc++;
	    break;
	    
	case 0x19:
	    //ORA -- Logical OR accumulator with memory [0x19]
	    //Absolute indexed with Y:   ORA addr16,Y    [3 bytes]
	    
	    a |= memory[idx_y ()];
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x1A:
	    break;
	    
	case 0x1B:
	    break;
	    
	case 0x1C:
	    break;
	    
	case 0x1D:
	    //ORA -- Logical OR accumulator with memory [0x1D]
	    //Absolute indexed with X:   ORA addr16,X [3 bytes]
	    
	    a |= memory[idx_x ()];
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x1E:
	    //ASL -- Arithmetic Shift left  [0x1E]
	    //Absolute indexed by X:    ASL addr16,X  [3 bytes]
	    
	    addr = idx_x ();
	    
	    set_carry_msb (memory[addr]);
	    
	    
	    memory[addr] <<= 1;
	    
	    set_ar_flags (memory[addr]);
	    
	    break;
	    
	case 0x1F:
	    break;
	    
	case 0x20:
	    //JSR -- Jump to Subroutine
	    
	    pc++; pc++;
	    push((byte)((pc >> 8) & 0xFF)); //push high byte
	    push((byte)(pc & 0xFF));    //low byte
	    pc = (memory[pc - 1] & 0xFF) + ((memory[pc] & 0xFF) << 8);
	    
	    break;
	    
	case 0x21:
	    //AND - Logical AND memory with accumulator [0x21]
	    //Pre-indexed indirect addressing mode:   AND (addr,X)   [2 bytes]
	    
	    
	    a &= memory[pre_idx ()];
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x22:
	    break;
	    
	case 0x23:
	    break;
	    
	case 0x24:
	    //BIT - Bit test
	    //Zero page direct addressing mode:  BIT addr [2 bytes]
	    
	    
	    addr = direct_zp ();
	    
	    temp = memory[addr];
	    
	    flags[OVERFLOW] = (temp & 64) != 0;
	    
	    flags[SIGN] = temp < 0;
	    
	    temp &= this.a;
	    
	    flags[ZERO] = (temp == 0);
	    
	    break;
	    
	case 0x25:
	    //AND - Logical AND accumulator with memory
	    //Zero page direct addressing mode:   AND addr [2 bytes]
	    
	    a &= memory[direct_zp ()];
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x26:
	    //ROL - Rotate bits left through carry
	    //Zero page direct addressing mode:	ROL addr	[2 bytes]
	    
	    addr = direct_zp ();
	    
	    if (flags[CARRY]) temp = 1;
	    else temp = 0;
	    
	    set_carry_msb (memory[addr]);
	    
	    memory[addr] <<= 1;
	    memory[addr] += temp;
	    
	    set_ar_flags (memory[addr]);
	    
	    break;
	    
	case 0x27:
	    break;
	    
	case 0x28:
	    //PLP - Pull contents of status register from stack
	    //Inherent addressing mode (no operand)		[1 byte]
	    
	    pc++;
	    temp = pop();
	    
	    set_status (temp);
	    
	    flags[5] = false;	//bit 5 of status register is not used
	    
	    break;
	    
	case 0x29:
	    //AND - Logical AND accumulator with data
	    //Immediate addressing mode:	AND data	[2 bytes]
	    
	    a &= immediate ();
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x2A:
	    //ROL - Rotate bits left through carry
	    //Accumulator addressing mode:	ROL A	[1 byte]
	    
	    pc++;
	    
	    if (flags[CARRY]) temp = 1;
	    else temp = 0;
	    
	    set_carry_msb (a);
	    
	    a <<= 1;
	    
	    a += temp;
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x2B:
	    break;
	    
	case 0x2C:
	    //BIT - Bit test
	    //Absolute direct addressing method: BIT addr16 [3 bytes];
	    
	    addr = direct ();
	    
	    temp = (byte)(a & memory[addr]);
	    
	    flags[ZERO] = (temp == 0);
	    
	    flags[SIGN] = (memory[addr] < 0);
	    
	    flags[OVERFLOW] = ((memory[addr] & 64) != 0);
	    
	    break;
	    
	case 0x2D:
	    //AND - Logical AND accumulator with data
	    //Immediate addressing mode: AND data [2 bytes]
	    
	    a &= immediate ();
	    
	    set_ar_flags (a);
	    
	    break;
	    
	case 0x2E:
	    //ROL - Rotate bits left through carry
	    //Absolute direct addressing mode: ROL addr16 [3 bytes]
	    
	    addr = direct ();
	    if (flags[CARRY]) temp = 1;
	    else temp = 0;
	    
	    set_carry_msb (memory[addr]);
	    
	    memory[addr] <<= 1;
	    memory[addr] += temp;
	    
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0x2F:
	    break;
	    
	case 0x30:
	    //BMI - Branch if minus (if sign flag [bit 7] is set
	    //BMI disp [2 bytes]
	    
	    pc++;
	    temp = memory[pc];
	    pc++;
	    if (flags[SIGN]) pc += temp;
	    break;
	    
	case 0x31:
	    //AND - Logical AND accumulator with memory
	    //Post-indexed indirect addressing mode: AND (addr),Y [2 bytes]
	    
	    addr = post_idx ();
	    a &= memory[addr];
	    set_ar_flags (a);
	    break;
	    
	case 0x32:
	    break;
	    
	case 0x33:
	    break;
	    
	case 0x34:
	    break;
	    
	    
	    
	case 0x35:
	    //AND - Logical AND accumulator with memory
	    //Zero page indexed by X: AND addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    a &= memory[addr];
	    set_ar_flags (a);
	    break;
	    
	case 0x36:
	    //ROL - Rotate bits left through carry
	    //Zero page indexed by X: ROL addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    
	    if (flags[CARRY]) temp = 1;
	    else temp = 0;
	    
	    set_carry_msb (memory[addr]);
	    
	    memory[addr] <<= 1;
	    memory[addr] += temp;
	    
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0x37:
	    break;
	    
	case 0x38:
	    //SEC - Set carry flag
	    //Implied: SEC [1 byte]
	    
	    pc++;
	    flags[CARRY] = true;
	    break;
	    
	case 0x39:
	    //AND - Logical AND accumulator with memory
	    //Absolute indexed by Y: AND addr16,Y [3 bytes]
	    
	    
	    addr = idx_y ();
	    a &= memory[addr];
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x3A:
	    break;
	    
	case 0x3B:
	    break;
	    
	case 0x3C:
	    break;
	    
	case 0x3D:
	    //AND - Logical AND accumulator with memory
	    //Absolute indexed by X: AND addr16,X [3 bytes]
	    
	    
	    addr = idx_x ();
	    
	    a &= memory[addr];
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x3E:
	    //ROL - Rotate bits left through carry
	    //Absolute indexed by X: ROL addr16,X [3 bytes]
	    
	    addr = idx_x ();
	    
	    
	    if (flags[CARRY]) temp = 1;
	    else temp = 0;
	    
	    set_carry_msb (memory[addr]);
	    
	    memory[addr] <<= 1;
	    memory[addr] += temp;
	    
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0x3F:
	    break;
	    
	case 0x40:
	    //RTI - Return from interrupt
	    //no addressing: RTI [1 byte]
	    
	    set_status(pop());
	    pc = (pop() & 0xFF) + ((pop() & 0xFF) << 8);
	    break;
	    
	case 0x41:
	    //EOR - Exclusive OR accumulator with memory
	    //Pre-indexed indirect addressing mode: EOR (addr,X) [2 bytes]
	    
	    
	    addr = pre_idx ();
	    
	    a ^= memory[addr];
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x42:
	    break;
	    
	case 0x43:
	    break;
	    
	case 0x44:
	    break;
	    
	case 0x45:
	    //EOR - Logical Exclusive OR accumulator with memory
	    //Zero page direct addressing mode: EOR addr [2 bytes];
	    
	    addr = direct_zp ();
	    a ^= memory[addr];
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x46:
	    //LSR - Logical shift right of accumulator or memory
	    //Zero page direct addressing mode: LSR [2 bytes]
	    
	    addr = direct_zp ();
	    
	    if ((memory[addr] & 1) != 0) flags[CARRY] = true;
	    else flags[CARRY] = false;
	    
	    memory[addr] >>>= 1;
	    
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0x47:
	    break;
	    
	case 0x48:
	    //PHA - Push accumulator onto stack
	    //implied addressing: PHA [1 byte]
	    
	    pc++;
	    push(a);
	    break;
	    
	case 0x49:
	    //EOR - Logical Exclusive OR accumulator with memory
	    //Immediate addressing mode: EOR data [2 bytes];
	    
	    a ^= immediate ();
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x4A:
	    //LSR - Logical shift right of accumulator or memory
	    //Register addressing mode: LSR A [1 byte]
	    
	    pc++;
	    
	    if((a & 1) != 0) flags[CARRY] = true;
	    else flags[CARRY] = false;
	    a >>>= 1;
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x4B:
	    break;
	    
	case 0x4C:
	    //JMP - Jump
	    //Absolute addressing mode: JMP label [3 bytes]
	    
	    pc = direct ();
	    
	    //		pc = (memory[pc] & 0xFF) + ((memory[++pc] & 0xFF) << 8);
	    
	    break;
	    
	case 0x4D:
	    //EOR - Logical Exclusive OR accmulator with memory
	    //Absolute direct addressing mode: EOR addr16 [3 bytes]
	    
	    addr = direct ();
	    a ^= memory[addr];
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x4E:
	    //LSR - Logical shift right of accumulator or memory
	    //Absolute direct addressing mode: LSR addr16 [3 bytes]
	    
	    
	    addr = direct ();
	    
	    if((memory[addr] & 1) != 0) flags[CARRY] = true;
	    else flags[CARRY] = false;
	    
	    memory[addr] >>>= 1;
	    
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0x4F:
	    break;
	    
	case 0x50:
	    //BVC - Branch if overflow (V) clear
	    //BVC disp [2 bytes]
	    
	    pc++;
	    if(! flags[OVERFLOW]) 
		{
		    pc += memory[pc];
		}
	    pc++;
	    
	    break;
	    
	case 0x51:
	    //EOR - logical exclusive-or accumulator with memory
	    //Post-indexed indirect addressing mode: EOR (addr),Y [2 bytes]
	    
	    addr = post_idx ();
	    a ^= memory[addr];
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x52:
	    break;
	    
	case 0x53:
	    break;
	    
	case 0x54:
	    break;
	    
	case 0x55:
	    //EOR - Logical exclusive-or accumulator with memory
	    //Zero page indexed by X: EOR addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    a ^= memory[addr];
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x56:
	    //LSR - Logical shift right bits of accumulator or memory
	    //Zero page indexed by x: LSR addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    
	    if((memory[addr] & 1) != 0) flags[CARRY] = true;
	    else flags[CARRY] = false;
	    
	    memory[addr] >>>= 1;
	    
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0x57:
	    break;
	    
	case 0x58:
	    //CLI - clear interrupt mask
	    //no operands
	    
	    pc++;
	    flags[INTERRUPT] = false;
	    break;
	    
	case 0x59:
	    //EOR - exclusive-or accumulator with memory
	    //Absolute indexed by Y: EOR addr16,Y [3 bytes]
	    
	    addr = idx_y ();
	    
	    a ^= memory[addr];
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x5A:
	    break;
	    
	case 0x5B:
	    break;
	    
	case 0x5C:
	    break;
	    
	case 0x5D:
	    //EOR - exclusive-or accumulator with memory
	    //Absolute indexed by X: EOR addr16,X [3 bytes]
	    
	    
	    addr = idx_x ();
	    
	    a ^= memory[addr];
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x5E:
	    //LSR - logical shift right bits of accumulator or memory
	    //Absolute indexed by X: LSR addr16,X [3 bytes]
	    
	    addr = idx_x ();
	    
	    memory[addr] >>>= 1;
	    
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0x5F:
	    break;
	    
	case 0x60:
	    //RTS - Return from subroutine
	    //inherent addressing : RTS [1 byte]
	    
	    addr = pop () & 0xFF;
	    addr += (pop () & 0xFF) << 8;
	    addr ++;
	    
	    pc = addr;
	    break;
	    
	case 0x61:
	    //ADC - Add with carry
	    //Pre-indexed indirect: ADC (addr,X) [2 bytes]
	    
	    addr = pre_idx ();
	    
	    if (flags[DECIMAL])
		{
		    adc_bcd (memory[addr]);
		}
	    else
		{
		    adc (memory[addr]);
		}
	    break;
	    
	case 0x62:
	case 0x63:
	case 0x64:
	    break;
	    
	case 0x65:
	    //ADC - Add with carry
	    //Zero page direct addressing: ADC addr [2 bytes]
	    
	    addr = direct_zp ();
	    
	    if (flags[DECIMAL])
		{
		    adc_bcd (memory[addr]);
		}
	    else
		{
		    adc (memory[addr]);
		}
	    
	    break;
	    
	case 0x66:
	    //ROR - Rotate memory right, through carry
	    //Zero page direct addressing: ROR addr [2 bytes]
	    
	    addr = direct_zp ();
	    
	    if (flags[CARRY]) temp = (byte)0x80;
	    else temp = 0;
	    
	    set_carry_lsb (memory[addr]);
	    
	    memory[addr] >>>= 1;
	    
	    memory[addr] += temp;
	    
	    set_ar_flags (memory[addr]);
	    
	    break;
	    
	case 0x67:
	    break;
	    
	case 0x68:
	    //PLA - Pull contents of accumulator from stack
	    //Inherent addressing:  PLA  [1 byte]
	    
	    pc++;
	    this.a = pop ();
	    set_ar_flags (this.a);
	    break;
	    
	case 0x69:
	    //ADC - Add to accumulator with carry
	    //Immediate addressing:  ADC data [2 bytes]
	    
	    temp = immediate ();
	    
	    if (flags[DECIMAL])
		{
		    adc_bcd (temp);
		}
	    else
		{
		    adc (temp);
		}
	    break;
	    
	case 0x6A:
	    //ROR - Rotate bits right through carry
	    //Register addressing:  ROR A [1 byte]
	    
	    pc++;
	    if (flags[CARRY]) temp = (byte)0x80;
	    else temp = 0;
	    
	    set_carry_lsb (this.a);
	    a >>>= 1;
	    a += temp;
	    
	    set_ar_flags (a);
	    break;
	    
	case 0x6B:
	    break;
	    
	case 0x6C:
	    //JMP - jump
	    //Indirect addressing: JMP (label) [3 bytes]
	    
	    pc = indirect ();
	    break;
	    
	case 0x6D:
	    //ADC - Add with carry
	    //Direct addressing: ADC addr16 [3 bytes]
	    
	    addr = direct ();
	    if (flags[DECIMAL])
		{
		    adc_bcd (memory[addr]);
		}
	    else
		{
		    adc (memory[addr]);
		}
	    break;
	    
	case 0x6E:
	    //ROR - Rotate bits right through carry
	    //Direct addressing: ROR addr16 [3 bytes]
	    
	    addr = direct ();
	    
	    if (flags[CARRY]) temp = (byte)(0x80);
	    else temp = 0;
	    
	    set_carry_lsb (memory[addr]);
	    memory[addr] >>>= 1;
	    memory[addr] += temp;
	    
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0x6F:
	    break;
	    
	case 0x70:
	    //BVS - Branch if overflow set
	    //Displacement operand:  BVS disp [2 bytes]
	    
	    pc++;
	    temp = memory[pc];
	    pc++;
	    
	    if (flags[OVERFLOW])
		{
		    pc += temp;
		}
	    break;
	    
	case 0x71:
	    //ADC - Add with carry
	    //Post-indexed indirect: ADC addr,Y [2 bytes]
	    
	    addr = post_idx ();
	    
	    if (flags[DECIMAL])
		{
		    adc_bcd (memory[addr]);
		}
	    else
		{
		    adc (memory[addr]);
		}
	    break;
	    
	case 0x72:
	case 0x73:
	case 0x74:
	    break;
	    
	case 0x75:
	    //ADC - Add with carry
	    //Zero-page, indexed with X: ADC addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    
	    if (flags[DECIMAL])
		{
		    adc_bcd (memory[addr]);
		}
	    else
		{
		    adc (memory[addr]);
		}
	    break;
	    
	case 0x76:
	    //ROR - Rotate bits right through carry
	    //Zero-page, indexed with X: ROR addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    
	    if (flags[CARRY]) temp = (byte)0x80;
	    else temp = 0;
	    
	    set_carry_lsb (memory[addr]);
	    memory[addr] >>>= 1;
	    memory[addr] += temp;
	    
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0x77:
	    break;
	    
	case 0x78:
	    //SEI - Set Interrupt mask (disable interrupts)
	    
	    pc++;
	    flags[INTERRUPT] = true;
	    break;
	    
	case 0x79:
	    //ADC - Add with carry
	    //Direct indexed with Y: ADC addr16,Y [3 bytes]
	    
	    addr = idx_y ();
	    
	    if (flags[DECIMAL])
		{
		    adc_bcd (memory[addr]);
		}
	    else
		{
		    adc (memory[addr]);
		}
	    break;
	    
	case 0x7A:
	case 0x7B:
	case 0x7C:
	    break;
	    
	case 0x7D:
	    //ADC - Add with carry
	    //Direct indexed with X: ADC addr16,X [3 bytes]
	    
	    addr = idx_x ();
	    
	    if (flags[DECIMAL])
		{
		    adc_bcd (memory[addr]);
		}
	    else
		{
		    adc (memory[addr]);
		}
	    break;
	    
	case 0x7E:
	    //ROR - Rotate bits right through carry
	    //Direct indexed with X: ROR addr16,X [3 bytes]
	    
	    addr = idx_x ();
	    
	    if (flags[CARRY]) temp = (byte)0x80;
	    else temp = 0;
	    
	    set_carry_lsb (memory[addr]);
	    memory[addr] >>>= 1;
	    memory[addr] += temp;
	    
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0x7F:
	case 0x80:
	    break;
	    
	case 0x81:
	    //STA - Store accumulator in memory
	    //Pre-indexed indirect: STA (addr,X) [2 bytes]
	    
	    addr = pre_idx ();
	    
	    memory[addr] = this.a;
	    break;
	    
	case 0x82:
	case 0x83:
	    break;
	    
	case 0x84:
	    //STY - Store Y in memory
	    //Zero-page direct: STY addr [2 bytes]
	    
	    addr = direct_zp ();
	    
	    memory[addr] = this.y;
	    break;
	    
	case 0x85:
	    //STA - Store accumulator in memory
	    //Zero-page direct: STA addr [2 bytes]
	    
	    addr = direct_zp ();
	    
	    memory[addr] = this.a;
	    break;
	    
	case 0x86:
	    //STX - Store X in memory
	    //Zero-page direct: STX addr [2 bytes]
	    
	    addr = direct_zp ();
	    memory[addr] = this.x;
	    break;
	    
	case 0x87:
	    break;
	    
	case 0x88:
	    //DEY - Decrement Y
	    //Inherent: DEY [1 byte]
	    
	    pc++;
	    this.y--;
	    set_ar_flags (this.y);
	    break;
	    
	case 0x89:
	    break;
	    
	case 0x8A:
	    //TXA - Transfer X to A
	    //Inherent: TXA [1 byte]
	    
	    pc++;
	    this.a = this.x;
	    set_ar_flags (this.a);
	    break;
	    
	case 0x8B:
	    break;
	    
	case 0x8C:
	    //STY - Store Y in memory
	    //Direct addressing: STY addr16 [3 bytes]
	    
	    addr = direct ();
	    
	    memory[addr] = this.y;
	    break;
	    
	case 0x8D:
	    //STA - Store accumulator in memory
	    //Direct addressing: STA addr16 [3 bytes]
	    
	    addr = direct ();
	    memory[addr] = this.a;
	    break;
	    
	case 0x8E:
	    //STX - Store X in memory
	    //Direct addressing: STX addr16 [3 bytes]
	    
	    addr = direct ();
	    memory[addr] = this.x;
	    break;
	    
	case 0x8F:
	    break;
	    
	case 0x90:
	    //BCC - Branch if carry clear
	    //Displacement: BCC disp [2 bytes]
	    
	    pc++;
	    temp = memory[pc];
	    pc++;
	    
	    if (!flags[CARRY]) pc += temp;
	    break;
	    
	case 0x91:
	    //STA - Store accumulator in memory
	    //Post-indexed indirect: STA (addr),Y [2 bytes]
	    
	    addr = post_idx ();
	    memory[addr] = this.a;
	    break;
	    
	case 0x92:
	case 0x93:
	    break;
	    
	case 0x94:
	    //STY - Store Y in memory
	    //Zero-page indexed with X: STY addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    memory[addr] = this.y;
	    break;
	    
	case 0x95:
	    //STA - Store accumulator in memory
	    //Zero-page indexed with X: STA addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    memory[addr] = this.a;
	    break;
	    
	case 0x96:
	    //STX - Store X in memory
	    //Zero-page indexed with Y: STX addr,Y [2 bytes]
	    
	    addr = idx_y_zp ();
	    memory[addr] = this.x;
	    break;
	    
	case 0x97:
	    break;
	    
	case 0x98:
	    //TYA - Transfer Y to Accumulator
	    //Inherent addressing: TYA [1 byte]
	    
	    pc++;
	    this.a = this.y;
	    set_ar_flags (a);
	    break;
	    
	case 0x99:
	    //STA - Store accumulator in memory
	    //Direct indexed with Y: STA addr16,Y [3 bytes]
	    
	    addr = idx_y ();
	    memory[addr] = this.a;
	    break;
	    
	case 0x9A:
	    //TXS - Transfer X to S
	    
	    pc++;
	    this.sp = this.x;
	    break;
	    
	case 0x9B:
	case 0x9C:
	    break;
	    
	case 0x9D:
	    //STA - Store accumulator in memory
	    //Direct indexed with X: STA addr16,X [3 bytes]
	    
	    addr = idx_x ();
	    memory[addr] = this.x;
	    break;
	    
	case 0x9E:
	case 0x9F:
	    break;
	    
	case 0xA0:
	    //LDY - Load Y with something
	    //Immediate: LDY data [2 bytes]
	    
	    this.y = immediate ();
	    set_ar_flags (y);
	    break;
	    
	case 0xA1:
	    //LDA - Load Accumulator
	    //Pre-indexed indirect: LDA (addr,X) [2 bytes]
	    
	    addr = pre_idx ();
	    this.a = memory[addr];
	    set_ar_flags (a);
	    break;
	    
	case 0xA2:
	    //LDX - Load X
	    //Immediate: LDX data [2 bytes]
	    
	    this.x = immediate ();
	    set_ar_flags (x);
	    break;
	    
	case 0xA3:
	    break;
	    
	case 0xA4:
	    //LDY - Load Y
	    //Zero-page direct: LDY addr [2 bytes]
	    
	    addr = direct_zp ();
	    this.y = memory[addr];
	    set_ar_flags (this.y);
	    break;
	    
	case 0xA5:
	    //LDA - Load Accumulator
	    //Zero-page direct: LDA addr [2 bytes]
	    
	    addr = direct_zp ();
	    this.a = memory[addr];
	    set_ar_flags (a);
	    break;
	    
	case 0xA6:
	    //LDX - Load X
	    //Zero-page direct: LDX addr [2 bytes]
	    
	    addr = direct_zp ();
	    this.x = memory[addr];
	    set_ar_flags (x);
	    break;
	    
	case 0xA7:
	    break;
	    
	case 0xA8:
	    //TAY - Transfer Accumulator to Y
	    //Inherent: TAY [1 byte]
	    
	    pc++;
	    this.y = this.a;
	    set_ar_flags (a);
	    break;
	    
	case 0xA9:
	    //LDA - Load Accumulator
	    //Immediate: LDA data [2 bytes]
	    
	    this.a = immediate ();
	    set_ar_flags (a);
	    break;
	    
	case 0xB0:
	    //BCS - Branch if carry set
	    //Displacement: BCD disp [2 bytes]
	    
	    temp = immediate ();
	    if (flags[CARRY]) pc += temp;
	    break;
	    
	case 0xB1:
	    //LDA - Load Accumulator from memory
	    //Post-indexed indirect: LDA (addr),Y [2 bytes]
	    
	    addr = post_idx ();
	    this.a = memory[addr];
	    set_ar_flags (a);
	    break;
	    
	case 0xB2:
	case 0xB3:
	    break;
	    
	case 0xB4:
	    //LDY - Load Y from memory
	    //Zero-page indexed with X: LDY addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    this.y = memory[addr];
	    set_ar_flags (y);
	    break;
	    
	case 0xB5:
	    //LDA - Load accumulator from memory
	    //Zero-page indexed with X: LDA addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    this.a = memory[addr];
	    set_ar_flags (a);
	    break;
	    
	case 0xB6:
	    //LDX - Load X from memory
	    //Zero-page indexed with Y: LDX addr,Y [2 bytes]
	    
	    addr = idx_y_zp ();
	    this.x = memory[addr];
	    set_ar_flags (x);
	    break;
	    
	case 0xB7:
	    break;
	    
	case 0xB8:
	    //CLV - Clear Overflow
	    //Inherent: CLV [1 byte]
	    
	    pc++;
	    flags[OVERFLOW] = false;
	    break;
	    
	case 0xB9:
	    //LDA - Load accumulator from memory
	    //Direct indexed with y: LDA addr16,Y [3 bytes]
	    
	    addr = idx_y ();
	    this.a = memory[addr];
	    set_ar_flags (a);
	    break;
	    
	case 0xBA:
	    //TSX - Transfer stack pointer to X
	    //Inherent: TSX [1 byte]
	    
	    pc++;
	    this.x = this.sp;
	    set_ar_flags (x);
	    break;
	    
	case 0xBB:
	    break;
	    
	case 0xBC:
	    //LDY - Load Y from memory
	    //Direct indexed with X: LDY addr16,X [3 bytes]
	    
	    addr = idx_x ();
	    this.y = memory[addr];
	    set_ar_flags (y);
	    break;
	    
	case 0xBD:
	    //LDA - Load accumulator from memory
	    //Direct indexed with X: LDA addr16,X [3 bytes]
	    
	    addr = idx_x ();
	    this.a = memory[addr];
	    set_ar_flags (a);
	    break;
	    
	case 0xBE:
	    //LDX - Load X from memory
	    //Direct indexed with Y: LDX addr16,Y [3 bytes]
	    
	    addr = idx_y ();
	    this.x = memory[addr];
	    set_ar_flags (x);
	    break;
	    
	case 0xBF:
	    break;
	    
	case 0xC0:
	    //CPY - Compare Y with memory
	    //Immediate: CPY data [2 bytes]
	    
	    temp = immediate ();
	    cmp (this.y, temp);
	    
	    break;
	    
	case 0xC1:
	    //CMP - Compare A with memory
	    //Pre-indexed indirect: CMP (addr,X) [2 bytes]
	    
	    addr = pre_idx ();
	    cmp (this.a, memory[addr]);
	    
	    break;
	    
	case 0xC2:
	case 0xC3:
	    break;
	    
	case 0xC4:
	    //CPY - Compare Y with memory
	    //Zero-page direct: CPY addr [2 bytes]
	    
	    addr = direct_zp ();
	    cmp (this.y, memory[addr]);
	    
	    break;
	    
	case 0xC5:
	    //CMP - Compare Accumulator with memory
	    //Zero-page direct: CMP addr [2 bytes]
	    
	    addr = direct_zp ();
	    cmp (this.a, memory[addr]);
	    break;
	    
	case 0xC6:
	    //DEC - Decrement memory
	    //Zero-page direct: DEC addr [2 bytes]
	    
	    addr = direct_zp ();
	    memory[addr]--;
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0xC7:
	    break;
	    
	case 0xC8:
	    //INY - Increment Y
	    //Inherent: INY [1 byte]
	    
	    pc++;
	    this.y++;
	    set_ar_flags (y);
	    break;
	    
	case 0xC9:
	    //CMP - Compare Accumulator with stuff
	    //Immediate: CMP data [2 bytes]
	    
	    temp = immediate ();
	    cmp (this.a, temp);
	    break;
	    
	case 0xCA:
	    //DEX - Decrement X
	    //Inherent: DEX [1 byte]
	    
	    pc++;
	    this.x--;
	    set_ar_flags (this.x);
	    break;
	    
	case 0xCB:
	    break;
	    
	case 0xCC:
	    //CPY - Compare Y with stuff
	    //Direct: CPY addr16 [3 bytes]
	    
	    addr = direct ();
	    cmp (this.y, memory[addr]);
	    break;
	    
	case 0xCD:
	    //CMP - Compare accumulator with stuff
	    //Direct: CMP addr16 [3 bytes]
	    
	    addr = direct ();
	    cmp (this.a, memory[addr]);
	    break;
	    
	case 0xCE:
	    //DEC - Decrement memory
	    //Direct: DEC addr16 [3 bytes]
	    
	    addr = direct ();
	    memory[addr]--;
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0xCF:
	    break;
	    
	case 0xD0:
	    //BNE - Branch if not equal
	    //Displacement: BNE disp [2 bytes]
	    
	    temp = immediate ();
	    if (!flags[ZERO]) pc += temp;
	    break;
	    
	case 0xD1:
	    //CMP - Compare Accumulator with memory
	    //Post-indexed indirect: CMP (addr),Y [2 bytes]
	    
	    addr = post_idx ();
	    cmp (this.a, memory[addr]);
	    break;
	    
	case 0xD2:
	case 0xD3:
	case 0xD4:
	    break;
	    
	case 0xD5:
	    //CMP - Compare Accumulator with memory
	    //Zero-page indexed with X: CMP addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    cmp (this.a, memory[addr]);
	    break;
	    
	case 0xD6:
	    //DEC - Decrement Memory
	    //Zero-page indexed with X: DEC addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    memory[addr]--;
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0xD7:
	    break;
	    
	case 0xD8:
	    //CLD - Clear Decimal flag
	    //Inherent: CLD [1 byte]
	    
	    pc++;
	    flags[DECIMAL] = false;
	    break;
	    
	case 0xD9:
	    //CMP - Compare accumulator with memory
	    //Direct indexed with Y: CMP addr16,Y [3 bytes]
	    
	    addr = idx_y ();
	    cmp (this.a, memory[addr]);
	    break;
	    
	case 0xE0:
	    //CPX - Compare X with stuff
	    //Immediate: CPX data [2 bytes]
	    
	    temp = immediate ();
	    cmp (this.x, temp);
	    break;
	    
	case 0xE1:
	    //SBC - Subtract with carry (borrow)
	    //Pre-indexed indirect: SBC (addr,X) [2 bytes]
	    
	    addr = pre_idx ();
	    if (flags[DECIMAL])
		{
		    sbc_bcd (memory[addr]);
		}
	    else
		{
		    sbc (memory[addr]);
		}
	    break;
	    
	case 0xE2:
	case 0xE3:
	    break;
	    
	case 0xE4:
	    //CPX - Compare x with ...
	    //Zero-page direct:  CPX addr [2 bytes]
	    
	    addr = direct_zp ();
	    
	    cmp (x, memory[addr]);
	    break;
	    
	case 0xE5:
	    //SBC - Subtract with carry
	    //Zero-page direct: SBC addr [2 bytes]
	    
	    addr = direct_zp ();
	    if (flags[DECIMAL])
		{
		    sbc_bcd (memory[addr]);
		}
	    else
		{
		    sbc (memory[addr]);
		}
	    break;
	    
	case 0xE6:
	    //INC - Increment memory
	    //Zero-page direct:  INC addr [2 bytes]
	    
	    addr = direct_zp ();
	    memory[addr]++;
	    set_ar_flags(memory[addr]);
	    break;
	    
	case 0xE7:
	    break;
	    
	case 0xE8:
	    //INX - Increment X
	    //inherent: INX [1 byte]
	    
	    pc++;
	    x++;
	    set_ar_flags (x);
	    break;
	    
	case 0xE9:
	    //SBC - Subtract with carry
	    //Immediate: SBC data [2 bytes]
	    
	    temp = immediate ();
	    if (flags[DECIMAL])
		{
		    sbc_bcd (temp);
		}
	    else
		{
		    sbc (temp);
		}
	    break;
	    
	case 0xEA:
	    //NOP - No operation
	    //Inherent: NOP [1 byte]
	    
	    pc++;
	    break;
	    
	case 0xEB:
	    break;
	    
	case 0xEC:
	    //CPX - Compare X
	    //Direct: CPX addr16 [3 bytes]
	    
	    addr = direct ();
	    cmp (x, memory[addr]);
	    break;
	    
	case 0xED:
	    //SBC - Subtract with Carry
	    //Direct:  SBC addr16 [3 bytes]
	    
	    addr = direct ();
	    if (flags[DECIMAL])
		{
		    sbc_bcd (memory[addr]);
		}
	    else
		{
		    sbc (memory[addr]);
		}
	    break;
	    
	case 0xEE:
	    //INC - Increment memory
	    //Direct: INC addr16 [3 bytes]
	    
	    addr = direct ();
	    memory[addr]++;
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0xEF:
	    break;
	    
	case 0xF0:
	    //BEQ - Branch if Equal
	    //Displacement: BEQ disp [2 bytes]
	    
	    temp = immediate ();
	    if (flags[ZERO]) pc += temp;
	    break;
	    
	case 0xF1:
	    //SBC - Subtract with carry
	    //Post-indexed indirect:  SBC (addr),Y [2 bytes]
	    
	    addr = post_idx ();
	    if (flags[DECIMAL])
		{
		    sbc_bcd (memory[addr]);
		}
	    else
		{
		    sbc (memory[addr]);
		}
	    break;
	    
	case 0xF2:
	case 0xF3:
	case 0xF4:
	    break;
	    
	case 0xF5:
	    //SBC - Subtract with carry
	    //Zero-page indexed with x: SBC addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    if (flags[DECIMAL])
		{
		    sbc_bcd (memory[addr]);
		}
	    else
		{
		    sbc (memory[addr]);
		}
	    break;
	    
	case 0xF6:
	    //INC - Increment memory
	    //Zero-page indexed with x: INC addr,X [2 bytes]
	    
	    addr = idx_x_zp ();
	    memory[addr] ++;
	    
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0xF7:
	    break;
	    
	case 0xF8:
	    //SED - Set decimal flag
	    //Inherent: SED [1 byte]
	    
	    pc++;
	    flags[DECIMAL] = true;
	    break;
	    
	case 0xF9:
	    //SBC - Subtract with carry
	    //Direct indexed with Y: SBC addr16,Y [3 bytes]
	    
	    addr = idx_y ();
	    
	    if (flags[DECIMAL])
		{
		    sbc_bcd (memory[addr]);
		}
	    else
		{
		    sbc (memory[addr]);
		}
	    break;
	    
	case 0xFA:
	case 0xFB:
	case 0xFC:
	    break;
	    
	case 0xFD:
	    //SBC - Subtract with carry
	    //Direct indexed with X: SBC addr16,X [3 bytes]
	    
	    addr = idx_x ();
	    if (flags[DECIMAL])
		{
		    sbc_bcd (memory[addr]);
		}
	    else
		{
		    sbc (memory[addr]);
		}
	    break;
	    
	case 0xFE:
	    //INC - Increment memory
	    //Direct indexed with X: INC addr16,X [3 bytes]
	    
	    addr = idx_x ();
	    memory[addr]++;
	    set_ar_flags (memory[addr]);
	    break;
	    
	case 0xFF:
	    break;
	    
	}
    }
}


